defmodule Pangram do
  @doc """
  Determines if a word or sentence is a pangram.
  A pangram is a sentence using every letter of the alphabet at least once.

  Also supports infinite streams of characters.

  Returns a boolean.

    ## Examples

      iex> Pangram.pangram?("the quick brown fox jumps over the lazy dog")
      true

      iex> Pangram.pangram?(Stream.cycle(String.codepoints("the quick brown fox jumps over the lazy dog")))
      true

  """

  use Bitwise

  @char_bits ?a..?z
    |> Enum.to_list |> to_string |> String.codepoints
    |> Enum.zip(Stream.iterate(0b1, &(&1 * 0b10)))
    |> Enum.into(%{})

  @pangram @char_bits |> Enum.reduce(0, &((elem(&1, 1) ||| &2)))

  @spec pangram?(String.t, timeout :: nil | integer) :: boolean
  def pangram?(sentence, timeout \\ 5000)

  def pangram?(sentence, timeout) when is_binary(sentence) do
    enum_is_pangram?(String.splitter(sentence, ""), timeout)
  end

  def pangram?(enum, timeout), do: enum_is_pangram?(enum, timeout)

  defp enum_is_pangram?(enum, :infinity) do
    enum
    |> Stream.scan(0b0, &(
      &2 ||| (
        @char_bits
        |> Map.get(String.downcase(&1), 0)
      )
    ))
    |> Enum.find_index(&(&1 == @pangram))
  end

  defp enum_is_pangram?(enum, timeout) do
    task = Task.async(fn -> enum_is_pangram?(enum, :infinity) end)
    case Task.yield(task, timeout) || Task.shutdown(task) do
      {:ok, result} -> result
      nil -> false
    end
  end
end